import { SummaryObject } from './latency-monitor.js';
import { EventEmitter, Startable } from '@libp2p/interfaces';
import { PeerId } from '@libp2p/interfaces/peer-id';
import type { Connection } from '@libp2p/interfaces/connection';
import type { ConnectionManager } from '@libp2p/interfaces/registrar';
import type { Components } from '@libp2p/interfaces/components';
export interface ConnectionManagerEvents {
    'peer:connect': CustomEvent<PeerId>;
    'peer:disconnect': CustomEvent<PeerId>;
}
export interface ConnectionManagerInit {
    /**
     * The maximum number of connections allowed
     */
    maxConnections?: number;
    /**
     * The minimum number of connections to avoid pruning
     */
    minConnections?: number;
    /**
     * The max data (in and out), per average interval to allow
     */
    maxData?: number;
    /**
     * The max outgoing data, per average interval to allow
     */
    maxSentData?: number;
    /**
     * The max incoming data, per average interval to allow
     */
    maxReceivedData?: number;
    /**
     * The upper limit the event loop can take to run
     */
    maxEventLoopDelay?: number;
    /**
     * How often, in milliseconds, metrics and latency should be checked
     */
    pollInterval?: number;
    /**
     * How often, in milliseconds, to compute averages
     */
    movingAverageInterval?: number;
    /**
     * The value of the peer
     */
    defaultPeerValue?: number;
    /**
     * Should preemptively guarantee connections are above the low watermark
     */
    autoDial?: boolean;
    /**
     * How often, in milliseconds, it should preemptively guarantee connections are above the low watermark
     */
    autoDialInterval?: number;
}
/**
 * Responsible for managing known connections.
 */
export declare class DefaultConnectionManager extends EventEmitter<ConnectionManagerEvents> implements ConnectionManager, Startable {
    private readonly components;
    private readonly init;
    private readonly peerValues;
    private readonly connections;
    private started;
    private timer?;
    private readonly latencyMonitor;
    constructor(components: Components, init?: ConnectionManagerInit);
    isStarted(): boolean;
    /**
     * Starts the Connection Manager. If Metrics are not enabled on libp2p
     * only event loop and connection limits will be monitored.
     */
    start(): Promise<void>;
    /**
     * Stops the Connection Manager
     */
    stop(): Promise<void>;
    /**
     * Cleans up the connections
     */
    _close(): Promise<void>;
    /**
     * Sets the value of the given peer. Peers with lower values
     * will be disconnected first.
     */
    setPeerValue(peerId: PeerId, value: number): void;
    /**
     * Checks the libp2p metrics to determine if any values have exceeded
     * the configured maximums.
     *
     * @private
     */
    _checkMetrics(): Promise<void>;
    /**
     * Tracks the incoming connection and check the connection limit
     */
    onConnect(evt: CustomEvent<Connection>): Promise<void>;
    /**
     * Removes the connection from tracking
     */
    onDisconnect(evt: CustomEvent<Connection>): void;
    getConnectionMap(): Map<string, Connection[]>;
    getConnectionList(): Connection[];
    getConnections(peerId: PeerId): Connection[];
    /**
     * Get a connection with a peer
     */
    getConnection(peerId: PeerId): Connection | undefined;
    /**
     * Get all open connections with a peer
     */
    getAll(peerId: PeerId): Connection[];
    /**
     * If the event loop is slow, maybe close a connection
     */
    _onLatencyMeasure(evt: CustomEvent<SummaryObject>): void;
    /**
     * If the `value` of `name` has exceeded its limit, maybe close a connection
     */
    _checkMaxLimit(name: keyof ConnectionManagerInit, value: number): Promise<void>;
    /**
     * If we have more connections than our maximum, close a connection
     * to the lowest valued peer.
     */
    _maybeDisconnectOne(): Promise<void>;
}
//# sourceMappingURL=index.d.ts.map